// Ersatz fuer FS20 S3 Garagentorsender
// 
// Der FS20 S3 Sender hat nur drei Tasten: Taste1, 2 und 4
// Der Digitalteil wird durch einen AtTiny85 ersetzt
// Die Tasten wecken den Micro aus dem Tiefschlaf
//
// V1.0 - 30.09.2019
// Die Codes (Hauscode und Adresse) werden manuell im Code eingetragen, die resultierenden Bitmuster werden errechnet
//
// V1.1 - 01.10.2019
// Umstellung auf reine Zahlen zur Speicherung und Berechnung (weg von den Strings).
// Dazu bitRead() und bitWrite() per Templates erweitert, damit auch unsigned long longs (ULL) bearbeitet werden koennen.
// Template von Nick Gammon, https://forum.arduino.cc/index.php?topic=294936.msg2058664#msg2058664
// 
// V1.5 - 02.10.2019
// Programmentwicklung auf DigiSpark Hardware, dazu ein #define eingefuehrt um zw. AtTiny und Digispark umzuschalten
// Portierung auf AtTiny und Implementierung von Deep Sleep -> Batterie sparen
// 08.10.2019
// dazu disablen:
// AD-Wandler, Analogkomparator, USI (universal serial interface), Brownout detection, Watchdog, 
//   interne Spannungsreferenz, ... Erfolg: 0,4..0,7 micro Ampere im deep sleep Modus :-)
//
// V1.6 - 15.10.2019
// Taste4 ist nicht ausprogrammiert, kann aber den Micro aus dem Sleep aufwecken. Leider geht er dann nicht mehr
//   in den Sleep Mode -> bitSet(PCMSK, Taste4) nicht ausfuehren oder Taste4 den EIN-Befehl senden lassen oder 
//   bei Taste4 den Micro einfach wieder in Sleep schicken
//   Option 1 (IRQ fuer Taste4 sperren) bewirkt, dass nur noch der Querstrom durch den PullUp fliesst, solange 
//   die Taste gedrueckt ist, immerhin 87 micro Ampere.
//
// V2.0 - todo
// Die Codes (Hauscode und Adresse) werden, wie beim Original, mit den Tasten eingegeben und im EEPROM gespeichert.
// Dazu muss Taste4 dann ebenfalls abgefragt werden (Taste4 ist vor V2.0 nicht implementiert)
// -------------------------------------------------------------------------------------------------------------

//#include <eeprom.h>	// erst ab V2.0
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/wdt.h>

// defines fuer Arduino AtTiny
// ACHTUNG
//#define DIGISPARK		// auskommentieren, wenn fr AtTiny generiert werden soll
//
#ifndef DIGISPARK
// defines fuer AtTiny (nicht DigiSpark)
#define FSout 1			// hier kommt das serielle Signal Richtung HF-Teil raus
#define Taste1 4		// internen PullUp einschalten
#define Taste2 0		// internen PullUp einschalten
#define Taste4 2		// internen PullUp einschalten (Taste4 wird im Programm nicht ausgewertet, keine Funktion)
#define Reset 5			// bleibt frei, nicht verwendet
#define LED 3
#else
// Beim DigiSpark kann P1 nicht zur Abfrage der Taste1 verwendet werden, dieser Anschluss ist immer auf LOW (USB-Beschaltung?)
// Daher fuer die Entwicklung auf dem DigiSpark Taste1 auf P2
// defines fuer DigiSpark (nicht AtTiny)
#define FSout 3			// hier kommt das serielle Signal Richtung HF-Teil raus
#define Taste1 2		// internal PullUp einschalten
#define Taste2 1		// internal PullUp einschalten
#define Taste4 4		// internal PullUp einschalten (Taste4 wird im Programm nicht ausgewertet, keine Funktion)
#define Reset 5			// bleibt frei, nicht verwendet
#define LED 0			// Ausgang fr die LED
#endif

// -------------------------------------------------------------------------------------------------------------
// Aufbau eines FS20 Telegramms
//							 |   HC1   |   HC2  |         |  Adr   |
// Code					HC:    11   41   41   41	Adr:    11   41		Befehl: ??		(Werte Tastencodes)
//						HC:    00   30   30   30	Adr:    00   30		Befehl: ??		(Werte zur Basis 4)
//						HC:     0    C    C    C	Adr:     0    C		Befehl: 11		(Werte zur Basis 16 - HEX)
//						HC:  0000 1100 1100 1100	Adr:  0000 1100		Befehl: 11		(Werte zur Basis 2 - BIN)
// globale Variablen
// Code fuer EIN
//              Parity ->                      P        P        P        P        P
//                        Synchron     HC1     |HC2     |Adresse |Befehl  |Quersum |EOT
//       BitsToSendEIN = "00000000000010000110001100110000000110000001000101111101110";
// -------------------------------------------------------------------------------------------------------------

static char Synchron[] = "0000000000001";	// Synchronisierung 12x0, 1x1, festgelegt, nicht aendern!
// ----------------------------------------------------------------------------------------------------------------------------
// hier die eigenenen Codes eintragen (HC1/HC2 sind der Hauscode, Einzelheiten siehe Bedienungsanleitung des FS20 S3 Senders)
static char HC1[] = "1141";					// Tastencodes (Taste1, Taste2, Taste4 am FS20 S3 moeglich)
static char HC2[] = "4141";
static char Adresse[] = "1141";
// ----------------------------------------------------------------------------------------------------------------------------
static char EOT[] = "0";					// kein Tastencode, nur eine Null, festgelegt, nicht aendern!

static unsigned char Befehl = 17;			// Achtung! KEIN Tastencode, sondern Dezimalzahl

unsigned char hexWert;						// Rueckgabewert fuer Codewandlung in HEX

unsigned long long BitsToSendEIN;			// die zu sendenden Bits werden in einem ULL abgelegt
unsigned long long BitsToSendAUS;			// die zu sendenden Bits werden in einem ULL abgelegt
unsigned char Hauscode1 = 0;				// Hauscode1
unsigned char Hauscode2 = 0;				// Hauscode2
unsigned char AdressByte = 0;				// AdressByte
unsigned char Quersumme = 0;				// Quersumme
unsigned long long uBitString = 0;

bool Taste1aktiv = false;					// Merker ob Taste1 gedrueckt wurde
bool Taste2aktiv = false;					// Merker ob Taste2 gedrueckt wurde
bool Taste4aktiv = false;					// Merker ob Taste4 gedrueckt wurde

bool ParityBit = false;						// Das Paritybit muss fuer jedes Byte des Code berechnet werden

// -------------------------------------------------------------------------------------------------------------
// bitRead() und bitWrite() per Templates erweitert, damit auch unsigned long longs (ULL) bearbeitet werden koennen.
// According to Nick Gammon - https://forum.arduino.cc/index.php?topic=294936.msg2058664#msg2058664

// die Arduino Routinen "ausknipsen", 
#undef bitRead
#undef bitWrite
// neue Funktionalitaet fuer bitRead/bitWrite, verarbeiten auch ULL Werte
template <typename T> void bitWrite(T & value, byte bit, byte val);  // prototype
template <typename T> void bitWrite(T & value, byte bit, byte val)
{
	if (val)
		value |= (T(1) << bit);
	else
		value &= ~(T(1) << bit);
}
template <typename T> bool bitRead(T value, byte bit);  // prototype
template <typename T> bool bitRead(T value, byte bit)
{
	return  (value & ((T)1 << bit)) != 0;
}
// \bitRead() und bitWrite() per Templates erweitert, damit auch unsigned long longs (ULL) bearbeitet werden koennen.
// -------------------------------------------------------------------------------------------------------------

// Interrruptroutine fuer alle Tasten
ISR(PCINT0_vect)
{
	// Die Routine muss nichts tun, nur den Micro aufwecken
	// Nach wake up geht es im Programmcode nach "sleep()" weiter

	// PCI triggert auf jedem Pin und bei jeder Flanke
}

// -------------------------------------------------------------------------------------------------------------
void setup()
{
	// IOs - Richtung festlegen
	pinMode(FSout, OUTPUT);
	pinMode(LED, OUTPUT);
	pinMode(Taste1, INPUT_PULLUP); pinMode(Taste2, INPUT_PULLUP); pinMode(Taste4, INPUT_PULLUP);
	//pinMode(Reset, INPUT);				// unnoetig, ist default

	// Impulsstrings zusammensetzen
	ImpulsstringBasteln(1);					// Aufruf mit Parameter fuer EIN
	BitsToSendEIN = uBitString;				// String fuellen mit EIN = 1

	ImpulsstringBasteln(0);					// Aufruf mit Parameter fuer AUS
	BitsToSendAUS = uBitString;				// String fuellen mit AUS = 0

	// PCI triggert auf jedem Pin und bei jeder Flanke
	// - Pins festlegen (PCMSK) 
	// - Pegel nach IRQ testen um fallende Flanke zu erkennen (Pins haben PullUps -> aktiv low)
	// alle Pins auf Interrupt einstellen
	bitSet(GIMSK, PCIE);					// PCINT enable, bei jeder Flanke wird ein IRQ erzeugt
	// Ich verwende hier nicht die defines fr PCINT sondern die fuer die Tasten. 
	//  Beide Nomenklaturen sind numerisch identisch (0..4 -- 5(Reset) ausgenommen)
	bitSet(PCMSK, Taste1);					// IRQ fuer Taste1 aktivieren
	bitSet(PCMSK, Taste2);					// IRQ fuer Taste2 aktivieren
	// Taste4 aus der IRQ Verarbeitung ausschliessen, sonst wacht der Micro auf und geht nicht mehr in den Sleep Mode
	//bitSet(PCMSK, Taste4);					// IRQ fuer Taste4 aktivieren

	cli();						// Interrupts ausschalten um Watchdog zu disablen
	WDT_off();					// watchdog ausschalten
	//BOD_off();					// brownout detection ausschalten (ist schon per Fuse disabled)
	sei();						// Interrupts wieder einschalten 

	// Use the Power Down sleep mode
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);

	// verschiedene HW-Ressourcen im MCU Control Register MCUCR zusaetzlich explizit abschalten (Strom sparen)
	// Power reduction register
	bitSet(PRR, PRTIM1);		// power reduction Timer1
	//bitSet(PRR, PRTIM0);		// power reduction Timer0 - nicht abschalten, sonst funktioniert delay() und delaymicroseconds() nicht
	bitSet(PRR, PRUSI);			// power reduction USI hardware

	// Achtung Falle!
	// Um die Analogen Komponenten des Micro ausschalten zu koennen, muss das Power Reduction Bit des ADC AUSgeschaltet sein
	// Erst nach dem Abschalten von ADC (und Komparator) darf das Power Reduction Bit gesetzt (EINgeschaltet) werden.
	// Analog comparator module
	bitSet(ACSR, ACD);			// power off analog comparator
	bitClear(ADCSRA, ADEN);		// disable AD converter
	bitSet(PRR, PRADC);			// power reduction AD converter (hier Power Reduction einschalten)

	sleep_mode();				//gehe in sleep und warte auf einen Interrupt...

}

// -------------------------------------------------------------------------------------------------------------
void loop()
{	
	// der Micro faellt hier in den deep sleep mode und wartet auf einen Tastendruck
	// sleep_mode() und die Kombination sleep_enable(), sleep_cpu(), sleep_disable() sind funktional identisch
	//sleep_mode();					//gehe in sleep und warte auf einen Interrupt...
	//sleep_enable();
	//sleep_cpu();
	//sleep_disable();
	bitClear(PRR, PRTIM0);		// power reduction Timer0 darf nicht erfolgen, sonst gibt es kein delay()

	// die nachfolgende Mimik sendet den Impuls nur einmal je Tastendruck - Taste1
	if ((digitalRead(Taste1) == LOW) && (Taste1aktiv == false))
	{
		// Taste1 abfragen
		Taste1aktiv = true;										// Trigger merken

		for (unsigned char i = 0; i < 59; i++)					// Code 1. Mal senden
		{
			sendBit(bitRead(BitsToSendEIN, i));
		}
		delay(10);												// Pause 10 ms

		for (unsigned char i = 0; i < 59; i++)					// Code 2. Mal senden
		{
			sendBit(bitRead(BitsToSendEIN, i));
		}
		delay(10);												// Pause 10 ms

		digitalWrite(LED, 1);

		for (unsigned char i = 0; i < 59; i++)					// Code 3. Mal senden
		{
			//(BitsToSendEIN[i] == '1') ? sendBit(1) : sendBit(0);
			sendBit(bitRead(BitsToSendEIN, i));
		}
		// LED ausschalten
		digitalWrite(LED, 0);
		delay(150);												// Pause >110 ms

		// fuer ganz kurzen Impuls hier testweise sofort wieder ausschalten - Hinweis: funktioniert
		//for (unsigned char i = 0; i < 59; i++)
		//{
		//	sendBit(bitRead(BitsToSendAUS, i));
		//}
	}
	else if ((Taste1aktiv == true) && (digitalRead(Taste1) == HIGH))
	{
		// erst wieder in Sleep gehen, wenn die Taste losgelassen wurde 
		//  (verhindern, dass die Flanke beim Loslassen ebenfalls einen IRQ triggert)
		// der Micro faellt hier in den deep sleep mode und wartet auf einen Tastendruck
		sleep_mode();											//gehe in sleep und warte auf einen Interrupt...
		// nach Sleep
		sleep_disable();										// hier geht es nach Tastendruck weiter

		Taste1aktiv = false;									// Trigger loeschen
		// Tastenprellen wird durch die Interrupt-Verarbeitung automatisch ignoriert
		//delay(500);											// 1/2 s warten um Tastenprellen zu ueberspielen
	}

	// ----------------------------------------------------------------------------------------------------------
	// die nachfolgende Mimik sendet den Impuls nur einmal je Tastendruck - Taste2
	if ((digitalRead(Taste2) == LOW) && (Taste2aktiv == false))
	{
		// LED einschalten
		digitalWrite(LED, 1);

		Taste2aktiv = true;										// Trigger merken

		for (unsigned char i = 0; i < 59; i++)					// Code 1. Mal senden
		{
			sendBit(bitRead(BitsToSendAUS, i));
		}
		delay(10);												// Pause 10 ms

		for (unsigned char i = 0; i < 59; i++)					// Code 2. Mal senden
		{
			sendBit(bitRead(BitsToSendAUS, i));
		}
		delay(10);												// Pause 10 ms

		// LED einschalten (erst hier, damit die LED nur kurz aufblitzt)
		digitalWrite(LED, 1);

		for (unsigned char i = 0; i < 59; i++)					// Code 3. Mal senden
		{
			sendBit(bitRead(BitsToSendAUS, i));
		}
		// LED ausschalten
		digitalWrite(LED, 0);

		delay(150);												// Pause >110 ms
	}
	else if ((Taste2aktiv == true) && (digitalRead(Taste2) == HIGH))
	{
		// erst wieder in Sleep gehen, wenn die Taste losgelassen wurde 
		//  (verhindern, dass die Flanke beim Loslassen ebenfalls einen IRQ triggert)
		// der Micro faellt hier in den deep sleep mode und wartet auf einen Tastendruck
		sleep_mode();											//gehe in sleep und warte auf einen Interrupt...
		// nach Sleep
		sleep_disable();										// hier geht es nach Tastendruck weiter

		Taste2aktiv = false;									// Trigger loeschen
		// Tastenprellen wird durch die Interrupt-Verarbeitung automatisch ignoriert
		//delay(500);											// 1/2 s warten um Tastenprellen zu ueberspielen
	}

}

// -------------------------------------------------------------------------------------------------------------
// schaltet den Ausgang Richtung HF-Teil entsprechend an/aus, zeitcodiert
void sendBit(unsigned char bitToSend)
{
	static int microSecondsEINS = 596;							// xx Mikrosekunden fuer binaere EINS
	static int microSecondsNULL = 380;							// xx Mikrosekunden fuer binaere NULL

	digitalWrite(FSout, HIGH);									// setze Ausgang
	bitToSend == 1 ? delayMicroseconds(microSecondsEINS) : delayMicroseconds(microSecondsNULL);		// warte Hi/Lo-Zeit
	digitalWrite(FSout, LOW);									// loesche Ausgang
	bitToSend == 1 ? delayMicroseconds(microSecondsEINS) : delayMicroseconds(microSecondsNULL);		// warte Hi/Lo-Zeit
}

// -------------------------------------------------------------------------------------------------------------
// Routine wandelt den uebergebenen 4-stelligen Code zur Basis 4 in ein HEX Byte um
unsigned char TastenToHex(String code)
{
	char intermediateCode[] = "5555";		// Zwischenspeicher fuer Wandlung
	unsigned char firstNibble, secondNibble;
	unsigned char i, retVal;

	// der uebergebene Code wird zuerst zeichenweise um 1 erniedrigt
	for (i = 0; i < code.length(); i++)
	{
		intermediateCode[i] = code[i] - 1;
	}

	// jetzt in HEX-Nibbles zerlegen
	firstNibble = (((intermediateCode[0] - 0x30) * 4) + (intermediateCode[1] - 0x30));
	secondNibble = (((intermediateCode[2] - 0x30) * 4) + (intermediateCode[3] - 0x30));

	// jetzt zu einer Hexzahl zusammenfassen
	retVal = firstNibble * 16 + secondNibble;

	return retVal;
}

// -------------------------------------------------------------------------------------------------------------
// gibt 1 (ODD) oder 0 (EVEN) parity fuer ein Byte zurueck
int parity(unsigned char x)
{
	x = x ^ x >> 4;
	x = x ^ x >> 2;
	x = x ^ x >> 1;
	return x & 1;
}

// -------------------------------------------------------------------------------------------------------------
// setzt den Impulsstring fr das EIN oder das AUS-Kommando zusammen
void ImpulsstringBasteln(unsigned char code)
{
	unsigned char i;

	// Quersumme berechnen, Quersumme startet mit 6
	Quersumme = 0x06;						// Quersumme vorbelegen mit 6
	hexWert = TastenToHex(HC1);
	Hauscode1 = hexWert;					// Hauscode1 in HEX merken fuer Impulsstring basteln

	Quersumme += hexWert;					// Quersumme aufaddieren

	hexWert = TastenToHex(HC2);
	Hauscode2 = hexWert;					// Hauscode2 in HEX merken fuer Impulsstring basteln

	Quersumme += hexWert;					// Quersumme aufaddieren

	hexWert = TastenToHex(Adresse);
	AdressByte = hexWert;					// AdressByte in HEX merken fuer Impulsstring basteln

	Quersumme += hexWert;					// Quersumme aufaddieren
	// Unterscheidung zwischen EIN und AUS Befehl
	Quersumme += ((code == 1) ? 17 : 0);	// Quersumme aufaddieren
	Befehl = ((code == 1) ? 17 : 0);		// Befehl richtig codieren EIN=17/AUS=0

	// Impulsstring zusammensetzen
	uBitString = 0;														// ist ein ULL
	unsigned char bitPointer = 0;										// Zeiger auf das zu schreibende Bit in uBitString
	for (i = 0; i < 13; i++)											// Synchronisation in BitString
	{
		bitWrite(uBitString, bitPointer++, (Synchron[i] == '1') ? 1 : 0); // Synchron ist ein char Array
	}
	for (i = 0; i < 8; i++)												// Hauscode1 in BitString
	{
		bitWrite(uBitString, bitPointer++, bitRead(Hauscode1, 7 - i));
	}
	bitWrite(uBitString, bitPointer++, parity(Hauscode1) == 1 ? 1:0);	// Parity von Hauscode1

	for (i = 0; i < 8; i++)												// Hauscode2 in BitString
	{
		bitWrite(uBitString, bitPointer++, bitRead(Hauscode2, 7 - i));
	}
	bitWrite(uBitString, bitPointer++, parity(Hauscode2) == 1 ? 1 : 0);	// Parity von Hauscode2

	for (i = 0; i < 8; i++)												// AdressByte in BitString
	{
		bitWrite(uBitString, bitPointer++, bitRead(AdressByte, 7 - i));
	}
	bitWrite(uBitString, bitPointer++, parity(AdressByte) == 1 ? 1 : 0);// Parity von Adresse

	for (i = 0; i < 8; i++)												// Befehl in BitString
	{
		bitWrite(uBitString, bitPointer++, bitRead(Befehl, 7 - i));
	}
	bitWrite(uBitString, bitPointer++, parity(Befehl) == 1 ? 1 : 0);	// Parity von Befehl

	for (i = 0; i < 8; i++)												// Quersumme in BitString
	{
		bitWrite(uBitString, bitPointer++, bitRead(Quersumme, 7 - i));
	}
	bitWrite(uBitString, bitPointer++, parity(Quersumme) == 1 ? 1 : 0);	// Parity von Quersumme

	bitWrite(uBitString, bitPointer++, 0);								// EOT ist immer 0 (Null)
}
// -------------------------------------------------------------------------------------------------------------
// Watchdog timer ausschalten
void WDT_off(void)
{
	wdt_reset();
	/* Clear WDRF in MCUSR */
	MCUSR = 0x00;
	/* Write logical one to WDCE and WDE */
	// die nachfolgenden beiden Befehle muessen innerhalb von 4 Clocks ausgefuehrt werden
	WDTCR |= (1 << WDCE) | (1 << WDE);
	/* Turn off WDT */
	WDTCR = 0x00;
}
// -------------------------------------------------------------------------------------------------------------
// Brownout detection ausschalten
void BOD_off(void)
{
	/* Clear WDRF in MCUSR */
	MCUSR = 0x00;
	/* Write logical one to BODS and BODSE */
	WDTCR |= (1 << BODS) | (1 << BODSE);
	// die nachfolgenden drei Befehle muessen innerhalb von 4 Clocks ausgefuehrt werden
	WDTCR |= (1 << BODS);
	WDTCR &= ~(1 << BODSE);
	/* Turn off BOD */
	MCUCR |= (1 << BODS);
}
// -------------------------------------------------------------------------------------------------------------
